{
        Program Name : Cache
        Package Name : LB
        Purpose      : Cache control routines
        Institution  : Conroy & Conroy Co.
        Date Written : 18-Jun-1994
        Written By   : Alan Conroy
        Version      : 1.0

        Copyright (C) 1994 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

	  This unit provides classes and functions for handling caches.
}

unit Cache ;

interface

{$I EDEFINES.INC}

uses Collect, { TCollection }
{$IFDEF BP}
     Objects, { TCollection }
{$ENDIF}
     _Cache, { TSCache }
     Common ; { TObject }

type TCache_Statistics = record
			     Hits : longint ; { Total read hits }
                             Misses : longint ; { Read misses }
                             { Note: Hits + Misses = total reads }
                             Writes : longint ; { Total writes }
                             Turns : longint ; { Number of times blocks were thrown away to make room for more }
                             Replacements : longint ; { Number of write hits }
                         end ;

type TSHeap_Cache = object( TSCache )
                        private { Instance data... }
                            Limit : longint ; { Largest amount of data we can cache, in bytes (0 = unlimited) }
                            Current : longint ; { Amount of data currently cached, in bytes }
                            Blocks : TCollection ; { Collection of cache blocks }

                        private { Internal utility routines... }
                            procedure Trim ;

                        public { API... }
                            function Read_Cache( var Data ;
                                Address, Size : longint ) : longint ; virtual ;
                            { Reads up to Size bytes from the cache for the specified
                              address.  Returns the number of bytes actually read (if
                              0 then the data was not cached).  Cached data is copied
                              to Data. }

                            procedure Write_Cache( const Data ;
                                Address, Size : longint ) ; virtual ;
                            { Writes to cache up to Size bytes from Data. }

                            procedure Clear ; virtual ; { Clears cache }

                            procedure Initialize ; virtual ; { Initialize object }

                            procedure Terminate ; virtual ; { Terminate object }

                            procedure Set_Max_Size( Size : longint ) ; virtual ;
                            { Sets maximum cache size in bytes.  If this is more
                              than is currently cached, some cache is thrown away. }

                            function Get_Max_Size : longint ; virtual ;
                            { Gets current maximum cache size in bytes. }

                            function Get_Size : longint ; virtual ;
                            { Returns amount of currently cached data, in bytes. }

                            function Get_Stats( Index : longint ;
                                var Address : longint ) : longint ;
                                virtual ;
                            { Returns the size of the data at cached index Index
                              and sets Address of the address of the cached
                              data. }

                            function Lock( Address : longint ) : boolean ; virtual ;
                            { Locks the data starting at the specified address into
                              memory so that it is not thrown away.  Returns false if
                              the specified data is not in the cache. }

                            function Unlock( Address : longint ) : boolean ; virtual ;
                            { Unlocks the data starting at the specified address in
                              memory.  Returns false if the specified data is not in
                              the cache, or if it is not locked. }
                    end ; { TSHeap_Cache }
     PSHeap_Cache = ^TSHeap_Cache ;


type TSHeap_Cache64 = object( TSCache64 )
                          private { Instance data... }
                              Limit : longint ; { Largest amount of data we can cache, in bytes (0 = unlimited) }
                              Current : longint ; { Amount of data currently cached, in bytes }
                              Blocks : TCollection ; { Collection of cache blocks }

                          private { Internal utility routines... }
                              procedure Trim ;

                          public { API... }
                              function Read_Cache( var Data ;
                                  Address, Size : int64 ) : longint ; virtual ;
                              { Reads up to Size bytes from the cache for the specified
                                address.  Returns the number of bytes actually read (if
                                0 then the data was not cached).  Cached data is copied
                                to Data. }

                              procedure Write_Cache( const Data ;
                                  Address, Size : int64 ) ; virtual ;
                              { Writes to cache up to Size bytes from Data. }

                              procedure Clear ; virtual ; { Clears cache }

                              procedure Initialize ; virtual ; { Initialize object }

                              procedure Terminate ; virtual ; { Terminate object }

                              procedure Set_Max_Size( Size : longint ) ;
                                  virtual ;
                              { Sets maximum cache size in bytes.  If this is more
                                than is currently cached, some cache is thrown away. }

                              function Get_Max_Size : longint ; virtual ;
                              { Gets current maximum cache size in bytes. }

                              function Get_Size : longint ; virtual ;
                              { Returns amount of currently cached data, in bytes. }

                              function Get_Stats( Index : longint ;
                                  var Address : int64 ) : longint ;
                                  virtual ;
                              { Returns the size of the data at cached index
                                index and sets Address of the address of the
                                cached data. }

                              function Lock( Address : int64 ) : boolean ;
                                  virtual ;
                              { Locks the data starting at the specified address into
                                memory so that it is not thrown away.  Returns false if
                                the specified data is not in the cache. }

                              function Unlock( Address : int64 ) : boolean ;
                                  virtual ;
                              { Unlocks the data starting at the specified address in
                                memory.  Returns false if the specified data is not in
                                the cache, or if it is not locked. }
                      end ; { TSHeap_Cache64 }
     PSHeap_Cache64 = ^TSHeap_Cache64 ;


{ The following code provides a means of coordinating several different
  caches. }

type TCache_Handler = function( Siz, Code : longint ) : longint ;
{ Cache handlers are passed a size (Siz) to free.  If size is zero (0) then
  Code indicates a special operation to perform.  For now, the only defined
  code is 0, which indicates that the function should return the amount of
  memory currently allocated in its cache.  If Size is not zero, Code is
  ignored.  If size is negative, then the cache should be completly zeroed.
  Other functions may be added later. }

procedure Add_Handler( Routine : TCache_Handler; Priority : longint ) ;
procedure Delete_Handler( Routine : TCache_Handler ) ;
procedure Free_Cache( Siz, Priority : longint; Exclude : TCache_Handler ) ;

{$IFDEF Test}
procedure Test_Unit ;
{$ENDIF}

implementation

uses Compatib ; { MaxAvail }

type TCache_Block = record
                        Address : longint ; { Address being cached }
                        Data : pointer ; { Pointer to cached data }
                        Size : longint ; { Size of data at Data }
                        Locked : boolean ; { True if this block is locked into cache }
                    end ;
     PCache_Block = ^TCache_Block ;


{ THeap_Cache methods... }

{ Internal utility routines... }

procedure TSHeap_Cache.Trim ;

var Block : PCache_Block ;

begin
    if( Blocks.Count > 0 ) then
    begin
        Block := Blocks.At( 0 ) ;
        freemem( Block^.Data, Block^.Size ) ;
        Current := Current - Block^.Size ;
        freemem( Block, sizeof( Block^ ) ) ;
        Blocks.AtDelete( 0 ) ;
    end ;
end ;


{ API... }

{ NOTE: Currently this implementation requires that reads and writes always
  start on specific addresses (ie, no reads in the middle of previously written
  blocks, etc.) }
function TSHeap_Cache.Read_Cache( var Data ; Address, Size : longint ) : longint ;

var Block : PCache_Block ;
    Loop : integer ;

begin
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then
        begin
            if( Size > Block^.Size ) then
            begin
                Size := Block^.Size ;
            end ;
            move( Block^.Data^, Data, Size ) ;
            Read_Cache := Size ;
            inc( Statistics.Hits ) ;
            exit ;
        end ;
    end ;

    Read_Cache := 0 ; { Failure }
    inc( Statistics.Misses ) ;
end ;


procedure TSHeap_Cache.Write_Cache( const Data ; Address, Size : longint ) ;

var Block : PCache_Block ;
    Loop : integer ;

begin
    if( ( Size > Limit ) and ( Limit > 0 ) ) then
    begin
        exit ; { This block is too big for our limit }
    end ;
    inc( Statistics.Writes ) ;
    if( Limit > 0 ) then
    begin
        { Make room }
        if( Current + Size > Limit ) then
        begin
            inc( Statistics.Turns ) ;
        end ;
        while( Current + Size > Limit ) do
        begin
	    Trim ;
        end ;
    end ;

    { See if block already in cache... }
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then { Found in cache }
        begin
            Blocks.AtDelete( Loop ) ; { Remove from list }
            Current := Current - Block^.Size ;
            if( Size <> Block^.Size ) then { Size mismatch }
            begin
                freemem( Block^.Data, Block^.Size ) ;
                freemem( Block, sizeof( Block^ ) ) ;
                break ;
            end else
            begin
                Blocks.Insert( Block ) ; { Put back into cache at end of list }
                Current := Current + Size ;
                move( Data, Block^.Data^, Size ) ;
                exit ;
            end ;
        end ;
    end ;

    { Add to cache... }
{$IFDEF DELPHI}
    try
{$ENDIF}
        new( Block ) ;
        fillchar( Block^, sizeof( Block^ ), 0 ) ;
{$IFDEF DELPHI}
    except
        exit ;
    end ;
{$ENDIF}
{$IFDEF DELPHI}
    try
{$ENDIF}
        getmem( Block^.Data, Size ) ;
{$IFDEF DELPHI}
    except
        Block^.Data := nil ;
    end ;
{$ENDIF}
    if( Block^.Data = nil ) then { Out of memory }
    begin
        freemem( Block, sizeof( Block^ ) ) ;
        exit ;
    end ;
    Block^.Address := Address ;
    Block^.Size := Size ;
    move( Data, Block^.Data^, Size ) ;
    Blocks.Insert( Block ) ;
    Current := Current + Size ;
end ; { TSHeap_Cache.Write_Cache }


procedure TSHeap_Cache.Clear ;

begin
    while( Current > 0 ) do
    begin
        Trim ;
    end ;
end ;


procedure TSHeap_Cache.Initialize ;

begin
    Limit := 0 ;
    Current := 0 ;
    Blocks.Init( 0, 4 ) ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;
end ;


procedure TSHeap_Cache.Terminate ;

begin
    Clear ;
    Blocks.Done ;
end ;


procedure TSHeap_Cache.Set_Max_Size( Size : longint ) ;

begin
    Limit := Size ;
    if( Limit > 0 ) then
    begin
        while( Limit < Current ) do { Trim data to fit new limit }
        begin
            Trim ;
        end ;
    end ;
end ;


function TSHeap_Cache.Get_Max_Size : longint ;

begin
    Get_Max_Size := Limit ;
end ;


function TSHeap_Cache.Get_Size : longint ;

begin
    Get_Size := Current ;
end ;


function TSHeap_Cache.Get_Stats( Index : longint ;
    var Address : longint ) : longint ;

var Block : PCache_Block ;

begin
    Get_Stats := 0 ;
    Address := -1 ;
    if( Blocks.Count > Index ) then
    begin
        Block := Blocks.At( Index ) ;
        Get_Stats := Block^.Size ;
        Address := Block^.Address ;
    end ;
end ;


function TSHeap_Cache.Lock( Address : longint ) : boolean ;

begin
    Lock := False ; { Not currently supported }
end ;


function TSHeap_Cache.Unlock( Address : longint ) : boolean ;

begin
    Unlock := False ; { Not currently supported }
end ;



type TCache64_Block = record
                          Address : int64 ; { Address being cached }
                          Data : pointer ; { Pointer to cached data }
                          Size : longint ; { Size of data at Data }
                          Locked : boolean ; { True if this block is locked into cache }
                      end ;
     PCache64_Block = ^TCache_Block ;


{ TSHeap_Cache64 methods... }

{ Internal utility routines... }

procedure TSHeap_Cache64.Trim ;

var Block : PCache64_Block ;

begin
    if( Blocks.Count > 0 ) then
    begin
        Block := Blocks.At( 0 ) ;
        freemem( Block^.Data, Block^.Size ) ;
        Current := Current - Block^.Size ;
        freemem( Block, sizeof( Block^ ) ) ;
        Blocks.AtDelete( 0 ) ;
    end ;
end ;


{ API... }

{ NOTE: Currently this implementation requires that reads and writes always
  start on specific addresses (ie, no reads in the middle of previously written
  blocks, etc.) }
function TSHeap_Cache64.Read_Cache( var Data ; Address, Size : int64 ) : longint ;

var Block : PCache64_Block ;
    Loop : integer ;

begin
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then
        begin
            if( Size > Block^.Size ) then
            begin
                Size := Block^.Size ;
            end ;
            move( Block^.Data^, Data, Size ) ;
            Read_Cache := Size ;
            inc( Statistics.Hits ) ;
            exit ;
        end ;
    end ;

    Read_Cache := 0 ; { Failure }
    inc( Statistics.Misses ) ;
end ;


procedure TSHeap_Cache64.Write_Cache( const Data ; Address, Size : int64 ) ;

var Block : PCache64_Block ;
    Loop : integer ;

begin
    if( ( Size > Limit ) and ( Limit > 0 ) ) then
    begin
        exit ; { This block is too big for our limit }
    end ;
    inc( Statistics.Writes ) ;
    if( Limit > 0 ) then
    begin
        { Make room }
        if( Current + Size > Limit ) then
        begin
            inc( Statistics.Turns ) ;
        end ;
        while( Current + Size > Limit ) do
        begin
	    Trim ;
        end ;
    end ;

    { See if block already in cache... }
    for Loop := 0 to Blocks.Count - 1 do
    begin
        Block := Blocks.At( Loop ) ;
        if( Block^.Address = Address ) then { Found in cache }
        begin
            Blocks.AtDelete( Loop ) ; { Remove from list }
            Current := Current - Block^.Size ;
            if( Size <> Block^.Size ) then { Size mismatch }
            begin
                freemem( Block^.Data, Block^.Size ) ;
                freemem( Block, sizeof( Block^ ) ) ;
                break ;
            end else
            begin
                Blocks.Insert( Block ) ; { Put back into cache at end of list }
                Current := Current + Size ;
                move( Data, Block^.Data^, Size ) ;
                exit ;
            end ;
        end ;
    end ;

    { Add to cache... }
{$IFDEF DELPHI}
    try
{$ENDIF}
        new( Block ) ;
        fillchar( Block^, sizeof( Block^ ), 0 ) ;
{$IFDEF DELPHI}
    except
        exit ;
    end ;
{$ENDIF}
{$IFDEF DELPHI}
    try
{$ENDIF}
        getmem( Block^.Data, Size ) ;
{$IFDEF DELPHI}
    except
        Block^.Data := nil ;
    end ;
{$ENDIF}
    if( Block^.Data = nil ) then { Out of memory }
    begin
        freemem( Block, sizeof( Block^ ) ) ;
        exit ;
    end ;
    Block^.Address := Address ;
    Block^.Size := Size ;
    move( Data, Block^.Data^, Size ) ;
    Blocks.Insert( Block ) ;
    Current := Current + Size ;
end ; { TSHeap_Cache64.Write_Cache }


procedure TSHeap_Cache64.Clear ;

begin
    while( Current > 0 ) do
    begin
        Trim ;
    end ;
end ;


procedure TSHeap_Cache64.Initialize ;

begin
    Limit := 0 ;
    Current := 0 ;
    Blocks.Init( 0, 4 ) ;
    fillchar( Statistics, sizeof( Statistics ), 0 ) ;
end ;


procedure TSHeap_Cache64.Terminate ;

begin
    Clear ;
    Blocks.Done ;
end ;


procedure TSHeap_Cache64.Set_Max_Size( Size : longint ) ;

begin
    Limit := Size ;
    if( Limit > 0 ) then
    begin
        while( Limit < Current ) do { Trim data to fit new limit }
        begin
            Trim ;
        end ;
    end ;
end ;


function TSHeap_Cache64.Get_Max_Size : longint ;

begin
    Get_Max_Size := Limit ;
end ;


function TSHeap_Cache64.Get_Size : longint ;

begin
    Get_Size := Current ;
end ;


function TSHeap_Cache64.Get_Stats( Index : longint ;
    var Address : int64 ) : longint ;

var Block : PCache64_Block ;

begin
    Get_Stats := 0 ;
    Address := -1 ;
    if( Blocks.Count > Index ) then
    begin
        Block := Blocks.At( Index ) ;
        Get_Stats := Block^.Size ;
        Address := Block^.Address ;
    end ;
end ;


function TSHeap_Cache64.Lock( Address : int64 ) : boolean ;

begin
    Lock := False ; { Not currently supported }
end ;


function TSHeap_Cache64.Unlock( Address : int64 ) : boolean ;

begin
    Unlock := False ; { Not currently supported }
end ;



{ Cache handler... }

type PCache_Record = ^TCache_Record ;
     TCache_Record = record
			 Routine : TCache_Handler ;
			 Priority : longint ;
			 Next : PCache_Record ;
		     end ;

var PCache_Handlers : PCache_Record ; { Root pointer to cache list }


procedure Add_Handler( Routine : TCache_Handler; Priority : longint ) ;
{ This routine adds a cache handler to the cache chain.  Routine is a pointer
  to the cache handler and Priority is the relative priority of this cache.
  The cache list is kept in ascending order by priority to make our job
  easier later. }

var Current, Previous : PCache_Record ;

begin
    if( PCache_Handlers = nil ) then { If first one, then it is simple }
    begin
	New( PCache_Handlers ) ;
	PCache_Handlers^.Next := nil ;
	PCache_Handlers^.Priority := Priority ;
	PCache_Handlers^.Routine := Routine ;
	exit ;
    end ;

    { We must insert into list... }
    Current := PCache_Handlers ;
    Previous := nil ;
    while ( Current <> nil ) do { Until end of list }
    begin
	if( Current^.Priority > Priority ) then { Found insertion point }
	begin
	    if( Previous = nil ) then { Insert at beginning of list }
	    begin
		New( PCache_Handlers ) ;
		Previous := PCache_Handlers ;
	    end else
	    begin
		New( Previous^.Next ) ; { Do insertion }
		Previous := Previous^.Next ;
	    end ;
	    Previous^.Next := Current ;
	    Previous^.Priority := Priority ;
	    Previous^.Routine := Routine ;
	    exit ;
	end ;
	Previous := Current ;
	Current := Current^.Next ; { Go to next list element }
    end ;

    { We hit the end of the list, add to end. }
    New( Current ) ;
    Previous^.Next := Current ;
    Current^.Next := nil ;
    Current^.Priority := Priority ;
    Current^.Routine := Routine ;
end ;


procedure Delete_Handler( Routine : TCache_Handler ) ;
{ This routine removes a cache handler from the cache handler list. }

var Current, Previous : PCache_Record ;

begin
    Current := PCache_Handlers ;
    Previous := nil ;
    while( Current <> nil ) do { Until end of list }
    begin
	if( @Current^.Routine = @Routine ) then { Found it }
	begin
	    if( Previous = nil ) then { First one in list }
            begin
	        PCache_Handlers := Current^.Next
            end else
            begin
	        Previous^.Next := Current^.Next ;
            end ;
	    Dispose( Current ) ; { Free up memory }
	    exit ;
	end ;
	Previous := Current ;
	Current := Current^.Next ; { Try next entry }
    end ;
end ;


procedure Free_Cache( Siz, Priority : longint ; Exclude : TCache_Handler ) ;
{ This routine calls the appropriate cache handlers to free up Siz bytes
  of memory.  A specific handler can be excluded from this.  Calling this
  with Exclude of nil, priority of maxint, and Siz of -1, will zero all
  caches. }

var Current : PCache_Record ;
{$IFNDEF DELPHI}
    Dummy : longint ;
{$ENDIF}

begin
    Current := PCache_Handlers ;
    while( Current <> nil ) do { Go through entire list }
    begin
	if( ( MaxAvail >= Siz ) and ( Siz >= 0 ) ) then
        begin { Enough memory is now available and we are not zeroing }
            Current := nil ; { Terminate loop }
        end else
        begin
            if(
                ( Current^.Priority <= Priority )
                and
                ( @Current^.Routine <> @Exclude )
              ) then
            begin
{$IFNDEF DELPHI}
                Dummy :=
{$ENDIF}
                Current^.Routine( Siz, 0 ) ;
            end ;
            Current := Current^.Next ; { Follow link to next one }
        end ;
    end ; { while( Current <> nil ) }
end ;


{$IFDEF Test}
procedure Error ;

begin
    writeln( 'Error' ) ;
    halt ;
end ;


var Called : boolean = false ;
    Called1 : boolean = false ;
    Called2 : boolean = false ;
    Called3 : boolean = false ;

function Test_Cache_Handler( Siz, Code : longint ) : longint ;

begin
    Called := True ;
    Test_Cache_Handler := 0 ;
end ;


function Test_Cache_Handler1( Siz, Code : longint ) : longint ;

begin
    Called1 := True ;
    Test_Cache_Handler1 := 0 ;
end ;


function Test_Cache_Handler2( Siz, Code : longint ) : longint ;

begin
    Called2 := True ;
    Test_Cache_Handler2 := 0 ;
end ;


function Test_Cache_Handler3( Siz, Code : longint ) : longint ;

begin
    Called3 := True ;
    Test_Cache_Handler3 := 0 ;
end ;


procedure Test_Add_Handler ;

begin
    Add_Handler( @Test_Cache_Handler, 3 ) ;
    if( PCache_Handlers = nil ) then
    begin
        Error ;
    end ;
    Add_Handler( @Test_Cache_Handler1, 1 ) ;
    Add_Handler( @Test_Cache_Handler2, 2 ) ;
    Add_Handler( @Test_Cache_Handler3, 4 ) ;
    if( @PCache_Handlers^.Routine <> @Test_Cache_Handler1 ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Delete_Handler ;

begin
    Delete_Handler( @Test_Cache_Handler1 ) ;
    Delete_Handler( @Test_Cache_Handler ) ;
    Delete_Handler( @Test_Cache_Handler2 ) ;
    Delete_Handler( @Test_Cache_Handler3 ) ;
    if( PCache_Handlers <> nil ) then
    begin
        Error ;
    end ;
    Delete_Handler( @Test_Cache_Handler2 ) ;
    Free_Cache( 1, 0, nil ) ;
end ;


procedure Test_Free_Cache ;

begin
    Free_Cache( 1, 0, nil ) ;
    Free_Cache( -1, 10, nil ) ;
    if( not ( Called and Called1 and Called2 and Called3 ) ) then
    begin
        Error ;
    end ;
end ;


procedure Test_Unit ;

begin
    Test_Add_Handler ;
    Test_Free_Cache ;
    Test_Delete_Handler ;
end ;
{$ENDIF}


begin
    PCache_Handlers := nil ; { vStart with no handlers loaded }
end.
